/*
Copyright © 2023 Philipp Wolfer <phw@uploadedlobster.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package cmd

import (
	"errors"
	"fmt"

	"github.com/manifoldco/promptui"
	"github.com/spf13/cobra"
	"go.uploadedlobster.com/scotty/internal/auth"
	"go.uploadedlobster.com/scotty/internal/backends"
	"go.uploadedlobster.com/scotty/internal/cli"
	"go.uploadedlobster.com/scotty/internal/config"
	"go.uploadedlobster.com/scotty/internal/i18n"
)

var serviceAddCmd = &cobra.Command{
	Use:   "add",
	Short: "Add a service configuration",
	Long:  `Interactively add a service to the configuration file.`,
	Run: func(cmd *cobra.Command, args []string) {
		// Select backend
		backend, err := cli.SelectBackend("")
		cobra.CheckErr(err)

		// Set service name
		prompt := promptui.Prompt{
			Label:   i18n.Tr("Service name"),
			Default: backend,
			Validate: func(s string) error {
				_, err := config.GetService(s)
				if err == nil {
					return errors.New(i18n.Tr("a service with this name already exists"))
				}
				return config.ValidateKey(s)
			},
		}
		name, err := prompt.Run()
		cobra.CheckErr(err)

		// Prepare service config
		service := config.ServiceConfig{
			Name:         name,
			Backend:      backend,
			ConfigValues: make(map[string]any),
		}

		// Additional options
		service, err = cli.PromptExtraOptions(service)
		cobra.CheckErr(err)

		// Save the service config
		err = service.Save()
		cobra.CheckErr(err)
		fmt.Println(i18n.Tr("Saved service %v using backend %v", service.Name, service.Backend))

		// Check whether authentication is required
		err = promptForAuth(service)
		cobra.CheckErr(err)
	},
}

func init() {
	serviceCmd.AddCommand(serviceAddCmd)

	// Here you will define your flags and configuration settings.

	// Cobra supports Persistent Flags which will work for this command
	// and all subcommands, e.g.:
	// serviceAddCmd.PersistentFlags().String("foo", "", "A help for foo")

	// Cobra supports local flags which will only run when this command
	// is called directly, e.g.:
	// serviceAddCmd.Flags().BoolP("toggle", "t", false, "Help message for toggle")
}

func promptForAuth(service config.ServiceConfig) error {
	backend, err := backends.ResolveBackend[auth.OAuth2Authenticator](service)
	if err != nil {
		// No authentication required, return
		return nil
	}

	doAuth, err := cli.PromptYesNo(
		i18n.Tr("The backend %v requires authentication. Authenticate now?", service.Backend),
		true,
	)
	if err != nil {
		return err
	}
	if !doAuth {
		return nil
	}

	cli.AuthenticationFlow(service, backend)
	return nil
}
