/*
Copyright © 2023-2025 Philipp Wolfer <phw@uploadedlobster.com>

This file is part of Scotty.

Scotty is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.

Scotty is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Scotty. If not, see <https://www.gnu.org/licenses/>.
*/
package listenbrainz

import (
	"context"
	"fmt"
	"sort"
	"time"

	"go.uploadedlobster.com/mbtypes"
	"go.uploadedlobster.com/musicbrainzws2"
	"go.uploadedlobster.com/scotty/internal/config"
	"go.uploadedlobster.com/scotty/internal/i18n"
	"go.uploadedlobster.com/scotty/internal/models"
	"go.uploadedlobster.com/scotty/internal/similarity"
	"go.uploadedlobster.com/scotty/internal/version"
	"go.uploadedlobster.com/scotty/pkg/listenbrainz"
)

type ListenBrainzApiBackend struct {
	client          listenbrainz.Client
	mbClient        musicbrainzws2.Client
	username        string
	checkDuplicates bool
	existingMBIDs   map[mbtypes.MBID]bool
}

func (b *ListenBrainzApiBackend) Name() string { return "listenbrainz" }

func (b *ListenBrainzApiBackend) Options() []models.BackendOption {
	return []models.BackendOption{{
		Name:  "username",
		Label: i18n.Tr("User name"),
		Type:  models.String,
	}, {
		Name:  "token",
		Label: i18n.Tr("Access token"),
		Type:  models.Secret,
	}, {
		Name:  "check-duplicate-listens",
		Label: i18n.Tr("Check for duplicate listens on import (slower)"),
		Type:  models.Bool,
	}}
}

func (b *ListenBrainzApiBackend) InitConfig(config *config.ServiceConfig) error {
	b.client = listenbrainz.NewClient(config.GetString("token"), version.UserAgent())
	b.mbClient = *musicbrainzws2.NewClient(musicbrainzws2.AppInfo{
		Name:    version.AppName,
		Version: version.AppVersion,
		URL:     version.AppURL,
	})
	b.client.MaxResults = listenbrainz.MaxItemsPerGet
	b.username = config.GetString("username")
	b.checkDuplicates = config.GetBool("check-duplicate-listens", false)
	return nil
}

func (b *ListenBrainzApiBackend) StartImport() error  { return nil }
func (b *ListenBrainzApiBackend) FinishImport() error { return nil }

func (b *ListenBrainzApiBackend) ExportListens(ctx context.Context, oldestTimestamp time.Time, results chan models.ListensResult, progress chan models.TransferProgress) {
	startTime := time.Now()
	minTime := oldestTimestamp
	if minTime.Unix() < 1 {
		minTime = time.Unix(1, 0)
	}

	totalDuration := startTime.Sub(oldestTimestamp)
	p := models.TransferProgress{
		Export: &models.Progress{
			Total: int64(totalDuration.Seconds()),
		},
	}

	for {
		result, err := b.client.GetListens(ctx, b.username, time.Now(), minTime)
		if err != nil {
			p.Export.Abort()
			progress <- p
			results <- models.ListensResult{Error: err}
			return
		}

		count := len(result.Payload.Listens)
		if count == 0 {
			if minTime.Unix() < result.Payload.OldestListenTimestamp {
				minTime = time.Unix(result.Payload.OldestListenTimestamp, 0)
				totalDuration = startTime.Sub(minTime)
				p.Export.Total = int64(totalDuration.Seconds())
				continue
			} else {
				break
			}
		}

		// Set minTime to the newest returned listen
		minTime = time.Unix(result.Payload.Listens[0].ListenedAt, 0)
		remainingTime := startTime.Sub(minTime)

		listens := make(models.ListensList, 0, count)

		for _, listen := range result.Payload.Listens {
			if listen.ListenedAt > oldestTimestamp.Unix() {
				listens = append(listens, AsListen(listen))
			} else {
				// result contains listens older then oldestTimestamp
				break
			}
		}

		sort.Sort(listens)
		p.Export.TotalItems += len(listens)
		p.Export.Elapsed = int64(totalDuration.Seconds() - remainingTime.Seconds())
		progress <- p
		results <- models.ListensResult{Items: listens, OldestTimestamp: minTime}
	}

	results <- models.ListensResult{OldestTimestamp: minTime}
	p.Export.Complete()
	progress <- p
}

func (b *ListenBrainzApiBackend) ImportListens(ctx context.Context, export models.ListensResult, importResult models.ImportResult, progress chan models.TransferProgress) (models.ImportResult, error) {
	total := len(export.Items)
	p := models.TransferProgress{}.FromImportResult(importResult, false)
	for i := 0; i < total; i += listenbrainz.MaxListensPerRequest {
		listens := export.Items[i:min(i+listenbrainz.MaxListensPerRequest, total)]
		count := len(listens)
		if count == 0 {
			break
		}

		submission := listenbrainz.ListenSubmission{
			ListenType: listenbrainz.Import,
			Payload:    make([]listenbrainz.Listen, 0, count),
		}

		for _, l := range listens {
			if b.checkDuplicates {
				isDupe, err := b.checkDuplicateListen(ctx, l)
				p.Import.Elapsed += 1
				progress <- p
				if err != nil {
					return importResult, err
				} else if isDupe {
					count -= 1
					msg := i18n.Tr("Ignored duplicate listen %v: \"%v\" by %v (%v)",
						l.ListenedAt, l.TrackName, l.ArtistName(), l.RecordingMBID)
					importResult.Log(models.Info, msg)
					continue
				}
			}

			l.FillAdditionalInfo()
			listen := listenbrainz.Listen{
				ListenedAt: l.ListenedAt.Unix(),
				TrackMetadata: listenbrainz.Track{
					TrackName:      l.TrackName,
					ReleaseName:    l.ReleaseName,
					ArtistName:     l.ArtistName(),
					AdditionalInfo: l.AdditionalInfo,
				},
			}
			listen.TrackMetadata.AdditionalInfo["submission_client"] = version.AppName
			listen.TrackMetadata.AdditionalInfo["submission_client_version"] = version.AppVersion

			submission.Payload = append(submission.Payload, listen)
		}

		if len(submission.Payload) > 0 {
			_, err := b.client.SubmitListens(ctx, submission)
			if err != nil {
				return importResult, err
			}
		}

		if count > 0 {
			importResult.UpdateTimestamp(listens[count-1].ListenedAt)
		}
		importResult.ImportCount += count
		progress <- p.FromImportResult(importResult, false)
	}

	return importResult, nil
}

func (b *ListenBrainzApiBackend) ExportLoves(ctx context.Context, oldestTimestamp time.Time, results chan models.LovesResult, progress chan models.TransferProgress) {
	exportChan := make(chan models.LovesResult)
	p := models.TransferProgress{
		Export: &models.Progress{},
	}

	go b.exportLoves(ctx, oldestTimestamp, exportChan)
	for existingLoves := range exportChan {
		if existingLoves.Error != nil {
			p.Export.Abort()
			progress <- p
			results <- models.LovesResult{Error: existingLoves.Error}
			return
		}

		p.Export.TotalItems = existingLoves.Total
		p.Export.Total = int64(existingLoves.Total)
		p.Export.Elapsed += int64(len(existingLoves.Items))
		progress <- p
		results <- existingLoves
	}

	p.Export.Complete()
	progress <- p
}

func (b *ListenBrainzApiBackend) exportLoves(ctx context.Context, oldestTimestamp time.Time, results chan models.LovesResult) {
	offset := 0
	defer close(results)
	loves := make(models.LovesList, 0, 2*listenbrainz.MaxItemsPerGet)

out:
	for {
		result, err := b.client.GetFeedback(ctx, b.username, 1, offset)
		if err != nil {
			results <- models.LovesResult{Error: err}
			return
		}

		count := len(result.Feedback)
		if count == 0 {
			break out
		}

		for _, feedback := range result.Feedback {
			// Missing track metadata indicates that the recording MBID is no
			// longer available and might have been merged. Try fetching details
			// from MusicBrainz.
			if feedback.TrackMetadata == nil {
				track, err := b.lookupRecording(ctx, feedback.RecordingMBID)
				if err == nil {
					feedback.TrackMetadata = track
				}
			}

			love := AsLove(feedback)
			if love.Created.After(oldestTimestamp) {
				loves = append(loves, love)
			} else {
				break out
			}
		}

		offset += listenbrainz.MaxItemsPerGet
	}

	sort.Sort(loves)
	results <- models.LovesResult{
		Total: len(loves),
		Items: loves,
	}
}

func (b *ListenBrainzApiBackend) ImportLoves(ctx context.Context, export models.LovesResult, importResult models.ImportResult, progress chan models.TransferProgress) (models.ImportResult, error) {
	if len(b.existingMBIDs) == 0 {
		existingLovesChan := make(chan models.LovesResult)
		go b.exportLoves(ctx, time.Unix(0, 0), existingLovesChan)

		// TODO: Store MBIDs directly
		b.existingMBIDs = make(map[mbtypes.MBID]bool, listenbrainz.MaxItemsPerGet)

		for existingLoves := range existingLovesChan {
			if existingLoves.Error != nil {
				return importResult, existingLoves.Error
			}

			for _, love := range existingLoves.Items {
				b.existingMBIDs[love.RecordingMBID] = true
				// In case the loved MBID got merged the track MBID represents the
				// actual recording MBID.
				if love.Track.RecordingMBID != "" &&
					love.Track.RecordingMBID != love.RecordingMBID {
					b.existingMBIDs[love.Track.RecordingMBID] = true
				}
			}
		}
	}

	for _, love := range export.Items {
		recordingMBID := love.RecordingMBID
		if recordingMBID == "" {
			recordingMBID = love.Track.RecordingMBID
		}

		if recordingMBID == "" {
			lookup, err := b.client.Lookup(ctx, love.TrackName, love.ArtistName())
			if err == nil {
				recordingMBID = lookup.RecordingMBID
			}
		}

		if recordingMBID != "" {
			ok := false
			errMsg := ""
			if b.existingMBIDs[recordingMBID] {
				ok = true
			} else {
				resp, err := b.client.SendFeedback(ctx, listenbrainz.Feedback{
					RecordingMBID: recordingMBID,
					Score:         1,
				})
				ok = err == nil && resp.Status == "ok"
				if err != nil {
					errMsg = err.Error()
				} else {
					b.existingMBIDs[recordingMBID] = true
				}
			}

			if ok {
				importResult.UpdateTimestamp(love.Created)
				importResult.ImportCount += 1
			} else {
				msg := fmt.Sprintf("Failed import of \"%s\" by %s: %v",
					love.TrackName, love.ArtistName(), errMsg)
				importResult.Log(models.Error, msg)
			}
		} else {
			msg := fmt.Sprintf("Failed import of \"%s\" by %s: no recording MBID",
				love.TrackName, love.ArtistName())
			importResult.Log(models.Error, msg)
		}

		progress <- models.TransferProgress{}.FromImportResult(importResult, false)
	}

	return importResult, nil
}

var defaultDuration = time.Duration(3 * time.Minute)

const trackSimilarityThreshold = 0.9

func (b *ListenBrainzApiBackend) checkDuplicateListen(ctx context.Context, listen models.Listen) (bool, error) {
	// Find listens
	duration := listen.Duration
	if duration == 0 {
		duration = defaultDuration
	}
	minTime := listen.ListenedAt.Add(-duration)
	maxTime := listen.ListenedAt.Add(duration)
	candidates, err := b.client.GetListens(ctx, b.username, maxTime, minTime)
	if err != nil {
		return false, err
	}

	for _, c := range candidates.Payload.Listens {
		sim := similarity.CompareTracks(listen.Track, AsTrack(c.TrackMetadata))
		if sim >= trackSimilarityThreshold {
			return true, nil
		}
	}

	return false, nil
}

func (b *ListenBrainzApiBackend) lookupRecording(
	ctx context.Context, mbid mbtypes.MBID) (*listenbrainz.Track, error) {
	filter := musicbrainzws2.IncludesFilter{
		Includes: []string{"artist-credits"},
	}
	recording, err := b.mbClient.LookupRecording(ctx, mbid, filter)
	if err != nil {
		return nil, err
	}

	artistMBIDs := make([]mbtypes.MBID, 0, len(recording.ArtistCredit))
	for _, artist := range recording.ArtistCredit {
		artistMBIDs = append(artistMBIDs, artist.Artist.ID)
	}
	track := listenbrainz.Track{
		TrackName:  recording.Title,
		ArtistName: recording.ArtistCredit.String(),
		MBIDMapping: &listenbrainz.MBIDMapping{
			// In case of redirects this MBID differs from the looked up MBID
			RecordingMBID: recording.ID,
			ArtistMBIDs:   artistMBIDs,
		},
	}
	return &track, nil
}

func AsListen(lbListen listenbrainz.Listen) models.Listen {
	listen := models.Listen{
		ListenedAt: time.Unix(lbListen.ListenedAt, 0),
		UserName:   lbListen.UserName,
		Track:      AsTrack(lbListen.TrackMetadata),
	}
	return listen
}

func AsLove(f listenbrainz.Feedback) models.Love {
	recordingMBID := f.RecordingMBID
	track := f.TrackMetadata
	if track == nil {
		track = &listenbrainz.Track{}
	}
	love := models.Love{
		UserName:      f.UserName,
		RecordingMBID: recordingMBID,
		Created:       time.Unix(f.Created, 0),
		Track:         AsTrack(*track),
	}

	if love.Track.RecordingMBID == "" {
		love.Track.RecordingMBID = love.RecordingMBID
	}

	return love
}

func AsTrack(t listenbrainz.Track) models.Track {
	track := models.Track{
		TrackName:        t.TrackName,
		ReleaseName:      t.ReleaseName,
		ArtistNames:      []string{t.ArtistName},
		Duration:         t.Duration(),
		TrackNumber:      t.TrackNumber(),
		DiscNumber:       t.DiscNumber(),
		RecordingMBID:    t.RecordingMBID(),
		ReleaseMBID:      t.ReleaseMBID(),
		ReleaseGroupMBID: t.ReleaseGroupMBID(),
		ISRC:             t.ISRC(),
		AdditionalInfo:   t.AdditionalInfo,
	}

	if t.MBIDMapping != nil && len(track.ArtistMBIDs) == 0 {
		for _, artistMBID := range t.MBIDMapping.ArtistMBIDs {
			track.ArtistMBIDs = append(track.ArtistMBIDs, artistMBID)
		}
	}

	return track
}
