/*
Copyright © 2023-2025 Philipp Wolfer <phw@uploadedlobster.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package models

import (
	"iter"
	"strings"
	"time"

	"go.uploadedlobster.com/mbtypes"
)

type Entity string

const (
	Listens Entity = "listens"
	Loves   Entity = "loves"
)

type AdditionalInfo map[string]any

type Track struct {
	TrackName        string
	ReleaseName      string
	ArtistNames      []string
	TrackNumber      int
	DiscNumber       int
	Duration         time.Duration
	ISRC             mbtypes.ISRC
	RecordingMBID    mbtypes.MBID
	ReleaseMBID      mbtypes.MBID
	ReleaseGroupMBID mbtypes.MBID
	ArtistMBIDs      []mbtypes.MBID
	WorkMBIDs        []mbtypes.MBID
	Tags             []string
	AdditionalInfo   AdditionalInfo
}

func (t Track) ArtistName() string {
	return strings.Join(t.ArtistNames, ", ")
}

// Updates AdditionalInfo to have standard fields as defined by ListenBrainz
func (t *Track) FillAdditionalInfo() {
	if t.AdditionalInfo == nil {
		t.AdditionalInfo = make(AdditionalInfo, 5)
	}
	if t.RecordingMBID != "" {
		t.AdditionalInfo["recording_mbid"] = t.RecordingMBID
	}
	if t.ReleaseGroupMBID != "" {
		t.AdditionalInfo["release_group_mbid"] = t.ReleaseGroupMBID
	}
	if t.ReleaseMBID != "" {
		t.AdditionalInfo["release_mbid"] = t.ReleaseMBID
	}
	if len(t.ArtistMBIDs) > 0 {
		t.AdditionalInfo["artist_mbids"] = t.ArtistMBIDs
	}
	if len(t.WorkMBIDs) > 0 {
		t.AdditionalInfo["work_mbids"] = t.WorkMBIDs
	}
	if t.ISRC != "" {
		t.AdditionalInfo["isrc"] = t.ISRC
	}
	if t.TrackNumber != 0 {
		t.AdditionalInfo["tracknumber"] = t.TrackNumber
	}
	if t.DiscNumber != 0 {
		t.AdditionalInfo["discnumber"] = t.DiscNumber
	}
	if t.Duration != 0 {
		rounded := t.Duration.Round(time.Second)
		if t.Duration == rounded {
			t.AdditionalInfo["duration"] = int64(t.Duration.Seconds())
		} else {
			t.AdditionalInfo["duration_ms"] = t.Duration.Milliseconds()
		}
	}
	if len(t.Tags) > 0 {
		t.AdditionalInfo["tags"] = t.Tags
	}
}

type Listen struct {
	Track
	ListenedAt       time.Time
	PlaybackDuration time.Duration
	UserName         string
}

type Love struct {
	Track
	Created       time.Time
	UserName      string
	RecordingMBID mbtypes.MBID
	RecordingMSID mbtypes.MBID
}

type ListensList []Listen

// Returns a new ListensList with only elements that are newer than t.
func (l ListensList) NewerThan(t time.Time) ListensList {
	result := make(ListensList, 0, len(l))
	for _, item := range l {
		if item.ListenedAt.After(t) {
			result = append(result, item)
		}
	}
	return result
}

func (l ListensList) Len() int {
	return len(l)
}

func (l ListensList) Less(i, j int) bool {
	return l[j].ListenedAt.After(l[i].ListenedAt)
}

func (l ListensList) Swap(i, j int) {
	l[i], l[j] = l[j], l[i]
}

type LovesList []Love

func (l LovesList) Len() int {
	return len(l)
}

func (l LovesList) Less(i, j int) bool {
	return l[j].Created.After(l[i].Created)
}

func (l LovesList) Swap(i, j int) {
	l[i], l[j] = l[j], l[i]
}

type ExportResult[T LovesList | ListensList] struct {
	Items           T
	Total           int
	OldestTimestamp time.Time
	Error           error
}

type ListensResult ExportResult[ListensList]

type LovesResult ExportResult[LovesList]

type LogEntryType string

const (
	Info    LogEntryType = "Info"
	Warning LogEntryType = "Warning"
	Error   LogEntryType = "Error"
)

type LogEntry struct {
	Type    LogEntryType
	Message string
}

type ImportResult struct {
	TotalCount    int
	ImportCount   int
	LastTimestamp time.Time
	ImportLog     []LogEntry

	// Error is only set if an unrecoverable import error occurred
	Error error
}

// Sets LastTimestamp to newTime, if newTime is newer than LastTimestamp
func (i *ImportResult) UpdateTimestamp(newTime time.Time) {
	if newTime.After(i.LastTimestamp) {
		i.LastTimestamp = newTime
	}
}

func (i *ImportResult) Update(from ImportResult) {
	i.TotalCount = from.TotalCount
	i.ImportCount = from.ImportCount
	i.UpdateTimestamp(from.LastTimestamp)
	i.ImportLog = append(i.ImportLog, from.ImportLog...)
}

func (i *ImportResult) Log(t LogEntryType, msg string) {
	i.ImportLog = append(i.ImportLog, LogEntry{
		Type:    t,
		Message: msg,
	})
}

type TransferProgress struct {
	Export *Progress
	Import *Progress
}

func (p TransferProgress) FromImportResult(result ImportResult, completed bool) TransferProgress {
	importProgress := Progress{
		Completed: completed,
	}.FromImportResult(result)
	p.Import = &importProgress
	return p
}

type Progress struct {
	TotalItems int
	Total      int64
	Elapsed    int64
	Completed  bool
	Aborted    bool
}

func (p Progress) FromImportResult(result ImportResult) Progress {
	p.Total = int64(result.TotalCount)
	p.Elapsed = int64(result.ImportCount)
	return p
}

func (p *Progress) Complete() {
	p.Elapsed = p.Total
	p.Completed = true
}

func (p *Progress) Abort() {
	p.Aborted = true
}

func IterExportProgress[T any](
	items []T, t *TransferProgress, c chan TransferProgress,
) iter.Seq2[int, T] {
	return iterProgress(items, t, t.Export, c, true)
}

func IterImportProgress[T any](
	items []T, t *TransferProgress, c chan TransferProgress,
) iter.Seq2[int, T] {
	return iterProgress(items, t, t.Import, c, false)
}

func iterProgress[T any](
	items []T, t *TransferProgress,
	p *Progress, c chan TransferProgress,
	autocomplete bool,
) iter.Seq2[int, T] {
	// Report progress in 1% steps
	steps := max(len(items)/100, 1)
	return func(yield func(int, T) bool) {
		for i, item := range items {
			if !yield(i, item) {
				return
			}
			p.Elapsed++
			if i%steps == 0 {
				c <- *t
			}
		}

		if autocomplete {
			p.Complete()
			c <- *t
		}
	}
}
