/*
Copyright © 2023-2025 Philipp Wolfer <phw@uploadedlobster.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

package jspf

import (
	"encoding/json"
	"fmt"
	"time"
)

// Represents a JSPF extension
type Extension any

// A map of JSPF extensions
type ExtensionMap map[string]Extension

// Parses the extension with the given ID and unmarshals it into "v".
// If the extensions is not found or the data cannot be unmarshalled,
// an error is returned.
func (e ExtensionMap) Get(id string, v any) error {
	ext, ok := e[id]
	if !ok {
		return fmt.Errorf("extension %q not found", id)
	}
	return unmarshalExtension(ext, v)
}

const (
	// The identifier for the MusicBrainz / ListenBrainz JSPF playlist extension
	MusicBrainzPlaylistExtensionID = "https://musicbrainz.org/doc/jspf#playlist"
	// The identifier for the MusicBrainz / ListenBrainz JSPF track extension
	MusicBrainzTrackExtensionID = "https://musicbrainz.org/doc/jspf#track"
)

// MusicBrainz / ListenBrainz JSPF track extension
//
// This extension allows storing additional metadata for playlists.
//
// See "https://musicbrainz.org/doc/jspf#playlist"
type MusicBrainzPlaylistExtension struct {
	// The ListenBrainz user who created this playlist.
	Creator string `json:"added_by,omitempty"`
	// Which ListenBrainz user was the playlist generated for? This is for music
	// recommendation bots generating playlists for users.
	CreatedFor string `json:"created_for,omitempty"`
	// Who are the ListenBrainz users who have access to edit this playlist?
	Collaborators []string `json:"collaborators,omitempty"`
	// This field identifies a playlist, using the identifier syntax,
	// from which this playlist was copied from.
	CopiedFrom string `json:"copied_from,omitempty"`
	// If the source playlist that this playlist has been copied from has been
	// deleted, this field will be set to true and the copied_from field will not
	// be returned.
	CopiedFromDeleted bool `json:"copied_from_deleted,omitempty"`
	// Indicates if this playlist is public or private.
	Public bool `json:"public,omitempty"`
	// The timestamp for when this playlist was last modified.
	LastModifiedAt time.Time `json:"last_modified_at,omitempty"`
	// This dict allows a playlist creator to submit additional track metadata
	// that may be used by playlist generation tools. The content of this field
	// is defined by the playlist generation tools and is beyond the scope of
	// this document.
	AdditionalMetadata map[string]any `json:"additional_metadata,omitempty"`
}

// MusicBrainz / ListenBrainz JSPF track extension
//
// This extension allows storing additional metadata for tracks.
//
// See "https://musicbrainz.org/doc/jspf#track"
type MusicBrainzTrackExtension struct {
	// The timestamp for when this track was added to the playlist.
	AddedAt time.Time `json:"added_at,omitempty"`
	// The ListenBrainz user who added this track.
	AddedBy string `json:"added_by,omitempty"`
	// The MusicBrainz ID URI for the release that contained this track.
	ReleaseIdentifier string `json:"release_identifier,omitempty"`
	// A list of MusicBrainz Artist URIs that identify the artist that are part
	// of the MusicBrainz artist credit id for this track.
	ArtistIdentifiers []string `json:"artist_identifiers,omitempty"`
	// This dict allows a playlist creator to submit additional track metadata
	// that may be used by playlist generation tools. The content of this field
	// is defined by the playlist generation tools and is beyond the scope of
	// this document.
	AdditionalMetadata map[string]any `json:"additional_metadata,omitempty"`
}

func unmarshalExtension(ext Extension, v any) error {
	asJson, err := json.Marshal(ext)
	if err != nil {
		return err
	}
	return json.Unmarshal(asJson, v)
}
